@@ -108,11 +108,6 @@ void PlaybackEngine::onRendererSynchronized(quint64 id, std::chrono::steady_cloc
     Q_ASSERT(m_renderers[QPlatformMediaPlayer::AudioStream]
              && m_renderers[QPlatformMediaPlayer::AudioStream]->id() == id);
 
-    if (m_timeController.positionFromTime(tp) < pos) {
-        // TODO: maybe check with an asset
-        qWarning() << "Unexpected synchronization " << m_timeController.positionFromTime(tp) - pos;
-    }
-
     m_timeController.sync(tp, pos);
 
     forEachExistingObject<Renderer>([&](auto &renderer) {
@@ -122,7 +117,7 @@ void PlaybackEngine::onRendererSynchronized(quint64 id, std::chrono::steady_cloc
 }
 
 void PlaybackEngine::setState(QMediaPlayer::PlaybackState state) {
-    if (!m_context)
+    if (!m_media.avContext())
         return;
 
     if (state == m_state)
@@ -202,7 +197,7 @@ PlaybackEngine::createRenderer(QPlatformMediaPlayer::TrackType trackType)
     switch (trackType) {
     case QPlatformMediaPlayer::VideoStream:
         return m_videoSink
-                ? createPlaybackEngineObject<VideoRenderer>(m_timeController, m_videoSink, getRotationAngle())
+                ? createPlaybackEngineObject<VideoRenderer>(m_timeController, m_videoSink, m_media.rotation())
                 : RendererPtr{ {}, {} };
     case QPlatformMediaPlayer::AudioStream:
         return m_audioOutput
@@ -309,7 +304,7 @@ void PlaybackEngine::recreateObjects()
 
 void PlaybackEngine::createObjectsIfNeeded()
 {
-    if (m_state == QMediaPlayer::StoppedState || !m_context)
+    if (m_state == QMediaPlayer::StoppedState || !m_media.avContext())
         return;
 
     for (int i = 0; i < QPlatformMediaPlayer::NTrackTypes; ++i)
@@ -368,7 +363,7 @@ void PlaybackEngine::createStreamAndRenderer(QPlatformMediaPlayer::TrackType tra
 
 std::optional<Codec> PlaybackEngine::codecForTrack(QPlatformMediaPlayer::TrackType trackType)
 {
-    const auto streamIndex = m_currentAVStreamIndex[trackType];
+    const auto streamIndex = m_media.currentStreamIndex(trackType);
     if (streamIndex < 0)
         return {};
 
@@ -377,7 +372,8 @@ std::optional<Codec> PlaybackEngine::codecForTrack(QPlatformMediaPlayer::TrackTy
     if (!result) {
         qCDebug(qLcPlaybackEngine)
                 << "Create codec for stream:" << streamIndex << "trackType:" << trackType;
-        auto maybeCodec = Codec::create(m_context->streams[streamIndex]);
+        auto maybeCodec =
+                Codec::create(m_media.avContext()->streams[streamIndex], m_media.avContext());
 
         if (!maybeCodec) {
             emit errorOccured(QMediaPlayer::FormatError,
@@ -399,13 +395,13 @@ bool PlaybackEngine::hasMediaStream() const
 
 void PlaybackEngine::createDemuxer()
 {
-    decltype(m_currentAVStreamIndex) streamIndexes = { -1, -1, -1 };
+    std::array<int, QPlatformMediaPlayer::NTrackTypes> streamIndexes = { -1, -1, -1 };
 
     bool hasStreams = false;
     forEachExistingObject<StreamDecoder>([&](auto &stream) {
         hasStreams = true;
         const auto trackType = stream->trackType();
-        streamIndexes[trackType] = m_currentAVStreamIndex[trackType];
+        streamIndexes[trackType] = m_media.currentStreamIndex(trackType);
     });
 
     if (!hasStreams)
@@ -413,9 +409,11 @@ void PlaybackEngine::createDemuxer()
 
     const PositionWithOffset positionWithOffset{ currentPosition(false), m_currentLoopOffset };
 
-    m_demuxer = createPlaybackEngineObject<Demuxer>(m_context.get(), positionWithOffset,
+    m_demuxer = createPlaybackEngineObject<Demuxer>(m_media.avContext(), positionWithOffset,
                                                     streamIndexes, m_loops);
 
+    connect(m_demuxer.get(), &Demuxer::packetsBuffered, this, &PlaybackEngine::buffered);
+
     forEachExistingObject<StreamDecoder>([&](auto &stream) {
         connect(m_demuxer.get(), Demuxer::signalByTrackType(stream->trackType()), stream.get(),
                 &StreamDecoder::decode);
@@ -457,24 +455,14 @@ void PlaybackEngine::deleteFreeThreads() {
         thr->wait();
 }
 
-bool PlaybackEngine::setMedia(const QUrl &media, QIODevice *stream)
+void PlaybackEngine::setMedia(MediaDataHolder media)
 {
-    stop();
-
-    // we should wait for objects deleting instead
-    // optimized solution might be implemented in the future.
-    deleteFreeThreads();
-
-    m_codecs = {};
-
-    if (auto error = recreateAVFormatContext(media, stream)) {
-        emit errorOccured(error->code, error->description);
-        return false;
-    }
+    Q_ASSERT(!m_media.avContext()); // Playback engine does not support reloading media
+    Q_ASSERT(m_state == QMediaPlayer::StoppedState);
+    Q_ASSERT(m_threads.empty());
 
+    m_media = std::move(media);
     updateVideoSinkSize();
-
-    return true;
 }
 
 void PlaybackEngine::setVideoSink(QVideoSink *sink)
@@ -528,12 +516,38 @@ qint64 PlaybackEngine::currentPosition(bool topPos) const {
                          : std::min(*pos, rendererPos);
     }
 
-    return boundPosition(pos ? *pos : m_timeController.currentPosition());
+    if (!pos)
+        pos = m_timeController.currentPosition();
+
+    return boundPosition(*pos - m_currentLoopOffset.pos);
+}
+
+qint64 PlaybackEngine::duration() const
+{
+    return m_media.duration();
+}
+
+bool PlaybackEngine::isSeekable() const { return m_media.isSeekable(); }
+
+const QList<MediaDataHolder::StreamInfo> &
+PlaybackEngine::streamInfo(QPlatformMediaPlayer::TrackType trackType) const
+{
+    return m_media.streamInfo(trackType);
+}
+
+const QMediaMetaData &PlaybackEngine::metaData() const
+{
+    return m_media.metaData();
+}
+
+int PlaybackEngine::activeTrack(QPlatformMediaPlayer::TrackType type) const
+{
+    return m_media.activeTrack(type);
 }
 
 void PlaybackEngine::setActiveTrack(QPlatformMediaPlayer::TrackType trackType, int streamNumber)
 {
-    if (!MediaDataHolder::setActiveTrack(trackType, streamNumber))
+    if (!m_media.setActiveTrack(trackType, streamNumber))
         return;
 
     m_codecs[trackType] = {};
@@ -593,8 +607,18 @@ void PlaybackEngine::updateVideoSinkSize(QVideoSink *prevSink)
 
     if (prevSink && prevSink->platformVideoSink())
         platformVideoSink->setNativeSize(prevSink->platformVideoSink()->nativeSize());
-    else if (auto size = metaData().value(QMediaMetaData::Resolution); size.isValid())
-        platformVideoSink->setNativeSize(size.value<QSize>());
+    else {
+        const auto streamIndex = m_media.currentStreamIndex(QPlatformMediaPlayer::VideoStream);
+        if (streamIndex >= 0) {
+            const auto context = m_media.avContext();
+            const auto stream = context->streams[streamIndex];
+            const auto pixelAspectRatio = av_guess_sample_aspect_ratio(context, stream, nullptr);
+            // auto size = metaData().value(QMediaMetaData::Resolution)
+            platformVideoSink->setNativeSize(
+                    qCalculateFrameSize({ stream->codecpar->width, stream->codecpar->height },
+                                        { pixelAspectRatio.num, pixelAspectRatio.den }));
+        }
+    }
 }
 
 qint64 PlaybackEngine::boundPosition(qint64 position) const
